@@ -1,6 +1,7 @@
 import warnings
 from functools import partial
-from itertools import (groupby,
+from itertools import (cycle,
+                       groupby,
                        repeat)
 from typing import (List,
                     Optional,
@@ -18,12 +19,17 @@ from .core.contracts import (is_contour_non_convex,
 from .hints import (Contour,
                     Coordinate,
                     Point,
+                    Polygon,
                     Polyline,
                     Segment,
                     Strategy)
-from .utils import (pack,
+from .utils import (constrict_convex_hull_size,
+                    pack,
                     to_concave_contour,
-                    to_convex_contour)
+                    to_convex_contour,
+                    to_convex_hull,
+                    to_polygon,
+                    to_strict_convex_hull)
 
 
 def points(x_coordinates: Strategy[Coordinate],
@@ -387,6 +393,7 @@ def contours(x_coordinates: Strategy[Coordinate],
     _validate_sizes(min_size, max_size, TRIANGLE_SIZE)
     if max_size is not None and max_size == TRIANGLE_SIZE:
         return triangular_contours(x_coordinates, y_coordinates)
+    min_size = max(min_size, TRIANGLE_SIZE)
     return (convex_contours(x_coordinates, y_coordinates,
                             min_size=min_size,
                             max_size=max_size)
@@ -591,6 +598,7 @@ def concave_contours(x_coordinates: Strategy[Coordinate],
     True
     """
     _validate_sizes(min_size, max_size, MIN_CONCAVE_CONTOUR_SIZE)
+    min_size = max(min_size, MIN_CONCAVE_CONTOUR_SIZE)
     return (strategies.lists(points(x_coordinates, y_coordinates),
                              min_size=min_size,
                              max_size=max_size,
@@ -681,25 +689,290 @@ def triangular_contours(x_coordinates: Strategy[Coordinate],
             .map(list))
 
 
+EMPTY_HOLES_SIZE = 0
+
+
+def polygons(x_coordinates: Strategy[Coordinate],
+             y_coordinates: Optional[Strategy[Coordinate]] = None,
+             *,
+             min_size: int = TRIANGLE_SIZE,
+             max_size: Optional[int] = None,
+             min_holes_size: int = EMPTY_HOLES_SIZE,
+             max_holes_size: Optional[int] = None,
+             min_hole_size: int = TRIANGLE_SIZE,
+             max_hole_size: Optional[int] = None) -> Strategy[Polygon]:
+    """
+    Returns a strategy for polygons.
+    Polygon is a pair of contour (called polygon’s border)
+    and possibly empty sequence of non-overlapping contours
+    which lie within the border (called polygon’s holes).
+
+    :param x_coordinates: strategy for vertices' x-coordinates.
+    :param y_coordinates:
+        strategy for vertices' y-coordinates,
+        ``None`` for reusing x-coordinates strategy.
+    :param min_size: lower bound for border size.
+    :param max_size: upper bound for border size, ``None`` for unbound.
+    :param min_holes_size: lower bound for holes count.
+    :param max_holes_size: upper bound for holes count, ``None`` for countless.
+    :param min_hole_size: lower bound for hole size.
+    :param max_hole_size: upper bound for hole size, ``None`` for unbound.
+
+    >>> from hypothesis import strategies
+    >>> from hypothesis_geometry import planar
+
+    For same coordinates' domain:
+
+    >>> min_coordinate, max_coordinate = -1., 1.
+    >>> coordinates_type = float
+    >>> coordinates = strategies.floats(min_coordinate, max_coordinate,
+    ...                                 allow_infinity=False,
+    ...                                 allow_nan=False)
+    >>> min_size, max_size = 5, 10
+    >>> min_holes_size, max_holes_size = 1, 4
+    >>> min_hole_size, max_hole_size = 3, 5
+    >>> polygons = planar.polygons(coordinates,
+    ...                            min_size=min_size,
+    ...                            max_size=max_size,
+    ...                            min_holes_size=min_holes_size,
+    ...                            max_holes_size=max_holes_size,
+    ...                            min_hole_size=min_hole_size,
+    ...                            max_hole_size=max_hole_size)
+    >>> polygon = polygons.example()
+    >>> isinstance(polygon, tuple)
+    True
+    >>> len(polygon) == 2
+    True
+    >>> border, holes = polygon
+    >>> isinstance(border, list)
+    True
+    >>> isinstance(holes, list)
+    True
+    >>> all(isinstance(hole, list) for hole in holes)
+    True
+    >>> min_size <= len(border) <= max_size
+    True
+    >>> min_holes_size <= len(holes) <= max_holes_size
+    True
+    >>> all(min_hole_size <= len(hole) <= max_hole_size for hole in holes)
+    True
+    >>> all(isinstance(vertex, tuple) for vertex in border)
+    True
+    >>> all(isinstance(vertex, tuple) for hole in holes for vertex in hole)
+    True
+    >>> all(len(vertex) == 2 for vertex in border)
+    True
+    >>> all(len(vertex) == 2 for hole in holes for vertex in hole)
+    True
+    >>> all(all(isinstance(coordinate, coordinates_type)
+    ...         for coordinate in vertex)
+    ...     for vertex in border)
+    True
+    >>> all(all(isinstance(coordinate, coordinates_type)
+    ...         for coordinate in vertex)
+    ...     for hole in holes
+    ...     for vertex in hole)
+    True
+    >>> all(all(min_coordinate <= coordinate <= max_coordinate
+    ...         for coordinate in vertex)
+    ...     for vertex in border)
+    True
+    >>> all(all(min_coordinate <= coordinate <= max_coordinate
+    ...         for coordinate in vertex)
+    ...     for hole in holes
+    ...     for vertex in hole)
+    True
+
+    For different coordinates' domains:
+
+    >>> min_x_coordinate, max_x_coordinate = -1., 1.
+    >>> min_y_coordinate, max_y_coordinate = 10., 100.
+    >>> coordinates_type = float
+    >>> x_coordinates = strategies.floats(min_x_coordinate, max_x_coordinate,
+    ...                                   allow_infinity=False,
+    ...                                   allow_nan=False)
+    >>> y_coordinates = strategies.floats(min_y_coordinate, max_y_coordinate,
+    ...                                   allow_infinity=False,
+    ...                                   allow_nan=False)
+    >>> min_size, max_size = 5, 10
+    >>> min_holes_size, max_holes_size = 1, 4
+    >>> min_hole_size, max_hole_size = 3, 5
+    >>> polygons = planar.polygons(x_coordinates, y_coordinates,
+    ...                            min_size=min_size,
+    ...                            max_size=max_size,
+    ...                            min_holes_size=min_holes_size,
+    ...                            max_holes_size=max_holes_size,
+    ...                            min_hole_size=min_hole_size,
+    ...                            max_hole_size=max_hole_size)
+    >>> polygon = polygons.example()
+    >>> isinstance(polygon, tuple)
+    True
+    >>> len(polygon) == 2
+    True
+    >>> border, holes = polygon
+    >>> isinstance(border, list)
+    True
+    >>> isinstance(holes, list)
+    True
+    >>> all(isinstance(hole, list) for hole in holes)
+    True
+    >>> min_size <= len(border) <= max_size
+    True
+    >>> min_holes_size <= len(holes) <= max_holes_size
+    True
+    >>> all(min_hole_size <= len(hole) <= max_hole_size for hole in holes)
+    True
+    >>> all(isinstance(vertex, tuple) for vertex in border)
+    True
+    >>> all(isinstance(vertex, tuple) for hole in holes for vertex in hole)
+    True
+    >>> all(len(vertex) == 2 for vertex in border)
+    True
+    >>> all(len(vertex) == 2 for hole in holes for vertex in hole)
+    True
+    >>> all(all(isinstance(coordinate, coordinates_type)
+    ...         for coordinate in vertex)
+    ...     for vertex in border)
+    True
+    >>> all(all(isinstance(coordinate, coordinates_type)
+    ...         for coordinate in vertex)
+    ...     for hole in holes
+    ...     for vertex in hole)
+    True
+    >>> all(min_x_coordinate <= vertex_x <= max_x_coordinate
+    ...     and min_y_coordinate <= vertex_y <= max_y_coordinate
+    ...     for vertex_x, vertex_y in border)
+    True
+    >>> all(min_x_coordinate <= vertex_x <= max_x_coordinate
+    ...     and min_y_coordinate <= vertex_y <= max_y_coordinate
+    ...     for hole in holes
+    ...     for vertex_x, vertex_y in hole)
+    True
+    """
+    _validate_sizes(min_size, max_size, TRIANGLE_SIZE)
+    _validate_sizes(min_holes_size, max_holes_size, EMPTY_HOLES_SIZE,
+                    'min_holes_size', 'max_holes_size')
+    _validate_sizes(min_hole_size, max_hole_size, TRIANGLE_SIZE,
+                    'min_hole_size', 'max_hole_size')
+    min_size, min_hole_size = (max(min_size, TRIANGLE_SIZE),
+                               max(min_hole_size, TRIANGLE_SIZE))
+
+    def to_points_with_sizes(points: List[Point]
+                             ) -> Strategy[Tuple[List[Point], int, List[int]]]:
+        max_border_points_count = len(points) - min_inner_points_count
+        min_border_size = max(min_size, len(to_strict_convex_hull(points)))
+        max_border_size = (max_border_points_count
+                           if max_size is None
+                           else min(max_size, max_border_points_count))
+        return strategies.tuples(strategies.just(points),
+                                 strategies.integers(min_border_size,
+                                                     max_border_size),
+                                 to_holes_sizes(points))
+
+    def to_holes_sizes(points: List[Point]) -> Strategy[List[int]]:
+        max_inner_points_count = len(points) - len(to_convex_hull(points))
+        holes_size_scale = max_inner_points_count // min_hole_size
+        points_max_hole_size = (holes_size_scale
+                                if max_holes_size is None
+                                else min(max_holes_size, holes_size_scale))
+        return (strategies.integers(min_holes_size, points_max_hole_size)
+                .flatmap(partial(_to_holes_sizes,
+                                 min_hole_size=min_hole_size,
+                                 max_hole_size=max_inner_points_count))
+                if max_inner_points_count >= min_hole_size
+                else strategies.builds(list))
+
+    def _to_holes_sizes(holes_size: int,
+                        min_hole_size: int,
+                        max_hole_size: int) -> Strategy[List[int]]:
+        if not holes_size:
+            return strategies.builds(list)
+        max_hole_size -= holes_size * min_hole_size
+        max_holes_sizes = [min_hole_size] * holes_size
+        indices = cycle(range(holes_size))
+        while max_hole_size:
+            max_holes_sizes[next(indices)] += 1
+            max_hole_size -= 1
+        sizes_ranges = [range(min_hole_size, max_hole_size + 1)
+                        for max_hole_size in max_holes_sizes]
+        return (strategies.permutations([strategies.sampled_from(sizes_range)
+                                         for sizes_range in sizes_ranges])
+                .flatmap(pack(strategies.tuples))
+                .map(list))
+
+    def has_valid_sizes(polygon: Polygon) -> bool:
+        border, holes = polygon
+        return (_has_valid_size(border,
+                                min_size=min_size,
+                                max_size=max_size)
+                and _has_valid_size(holes,
+                                    min_size=min_holes_size,
+                                    max_size=max_holes_size)
+                and all(_has_valid_size(hole,
+                                        min_size=min_hole_size,
+                                        max_size=max_hole_size)
+                        for hole in holes))
+
+    min_inner_points_count = min_hole_size * min_holes_size
+
+    def has_valid_inner_points_count(points: List[Point]) -> bool:
+        return ((max_size is None
+                 or len(to_strict_convex_hull(points)) <= max_size)
+                and (len(points) - len(to_convex_hull(points))
+                     >= min_inner_points_count))
+
+    return (strategies.lists(
+            points(x_coordinates, y_coordinates),
+            min_size=min_size + min_inner_points_count,
+            max_size=(None
+                      if (max_size is None
+                          or max_holes_size is None
+                          or max_hole_size is None)
+                      else max_size + max_hole_size * max_holes_size),
+            unique=True)
+            .map(sorted)
+            .map(partial(constrict_convex_hull_size,
+                         max_size=max_size))
+            .filter(has_valid_inner_points_count)
+            .flatmap(to_points_with_sizes)
+            .map(pack(to_polygon))
+            .filter(has_valid_sizes))
+
+
 def _validate_sizes(min_size: int, max_size: Optional[int],
-                    min_expected_size: int) -> None:
+                    min_expected_size: int,
+                    min_size_name: str = 'min_size',
+                    max_size_name: str = 'max_size') -> None:
     if max_size is None:
         pass
     elif max_size < min_expected_size:
-        raise ValueError('Should have at least {expected} vertices, '
-                         'but requested {actual}.'
-                         .format(expected=min_expected_size,
-                                 actual=max_size))
-    elif min_size > max_size:
-        raise ValueError('`min_size` should not be greater than `max_size`, '
-                         'but found {min_size}, {max_size}.'
-                         .format(min_size=min_size,
+        raise ValueError('`{max_size_name}` '
+                         'should not be less than {min_expected_size}, '
+                         'but found {max_size}.'
+                         .format(max_size_name=max_size_name,
+                                 min_expected_size=min_expected_size,
                                  max_size=max_size))
+    elif min_size > max_size:
+        raise ValueError('`{min_size_name}` '
+                         'should not be greater than `{max_size_name}`, '
+                         'but found {min_size}, {max_size}.'
+                         .format(min_size_name=min_size_name,
+                                 max_size_name=max_size_name,
+                                 min_size=min_size,
+                                 max_size=max_size))
+    elif min_size < 0:
+        raise ValueError('`{min_size_name}` '
+                         'should not be less than 0, '
+                         'but found {min_size}.'
+                         .format(min_size_name=min_size_name,
+                                 min_size=min_size))
     if min_size < min_expected_size:
-        warnings.warn('`min_size` is expected to be not less than {expected}, '
-                      'but found {actual}.'
-                      .format(expected=min_expected_size,
-                              actual=min_size),
+        warnings.warn('`{min_size_name}` is expected to be '
+                      'not less than {min_expected_size}, '
+                      'but found {min_size}.'
+                      .format(min_size_name=min_size_name,
+                              min_expected_size=min_expected_size,
+                              min_size=min_size),
                       HypothesisWarning)
 
 
