@@ -3,6 +3,7 @@ from math import atan2
 from typing import (Callable,
                     Iterable,
                     List,
+                    Optional,
                     Sequence,
                     Tuple,
                     TypeVar)
@@ -17,7 +18,8 @@ from .core.utils import (Orientation,
                          to_orientation)
 from .hints import (Contour,
                     Coordinate,
-                    Point)
+                    Point,
+                    Polygon)
 
 
 def to_concave_contour(points: Sequence[Point]) -> Contour:
@@ -32,25 +34,16 @@ def to_concave_contour(points: Sequence[Point]) -> Contour:
         http://www.geosensor.net/papers/duckham08.PR.pdf
     """
     triangulation = triangular.delaunay(points)
-    boundary = triangular.to_boundary_edges(triangulation)
-    boundary_vertices = {edge.start for edge in boundary}
+    boundary_edges = triangular.to_boundary_edges(triangulation)
+    boundary_vertices = {edge.start for edge in boundary_edges}
 
     def is_mouth(edge: QuadEdge) -> bool:
         return edge.left_from_start.end not in boundary_vertices
 
     edges_neighbours = {edge: to_edge_neighbours(edge)
-                        for edge in boundary}
-
-    def edge_key(edge: QuadEdge) -> Sortable:
-        return to_squared_edge_length(edge), edge.start, edge.end
-
-    def to_squared_edge_length(edge: QuadEdge) -> Coordinate:
-        (start_x, start_y), (end_x, end_y) = edge.start, edge.end
-        delta_x, delta_y = start_x - end_x, start_y - end_y
-        return delta_x * delta_x + delta_y * delta_y
-
-    candidates = red_black.tree(*filter(is_mouth, boundary),
-                                key=edge_key)
+                        for edge in boundary_edges}
+    candidates = red_black.tree(*filter(is_mouth, boundary_edges),
+                                key=_edge_key)
     while candidates:
         edge = candidates.popmax()
         if not is_mouth(edge):
@@ -60,12 +53,101 @@ def to_concave_contour(points: Sequence[Point]) -> Contour:
         for neighbour in edges_neighbours.pop(edge):
             edges_neighbours[neighbour] = to_edge_neighbours(neighbour)
             candidates.add(neighbour)
-    result = [edge.start for edge in
-              triangular.to_boundary_edges(triangulation)]
+    result = [edge.start
+              for edge in triangular.to_boundary_edges(triangulation)]
     shrink_collinear_vertices(result)
     return result
 
 
+def to_polygon(points: Sequence[Point],
+               size: int,
+               holes_sizes: List[int]) -> Polygon:
+    triangulation = triangular.delaunay(points)
+    boundary_edges = triangular.to_boundary_edges(triangulation)
+    boundary_vertices = {edge.start for edge in boundary_edges}
+    inner_points = sorted(set(points) - boundary_vertices)
+    start = 0
+    holes = []
+    for hole_size in holes_sizes:
+        hole_points = inner_points[start:start + hole_size]
+        holes.append(to_concave_contour(hole_points))
+        boundary_vertices.update(hole_points)
+        start += hole_size
+
+    def is_mouth(edge: QuadEdge) -> bool:
+        return edge.left_from_start.end not in boundary_vertices
+
+    edges_neighbours = {edge: to_edge_neighbours(edge)
+                        for edge in boundary_edges}
+    candidates = red_black.tree(*filter(is_mouth, boundary_edges),
+                                key=_edge_key)
+    current_size = len(to_strict_convex_hull(points))
+    while current_size < size:
+        try:
+            edge = candidates.popmax()
+        except KeyError:
+            break
+        if not is_mouth(edge):
+            continue
+        current_size += 1
+        boundary_vertices.add(edge.left_from_start.end)
+        triangulation.delete(edge)
+        for neighbour in edges_neighbours.pop(edge):
+            edges_neighbours[neighbour] = to_edge_neighbours(neighbour)
+            candidates.add(neighbour)
+    border = [edge.start
+              for edge in triangular.to_boundary_edges(triangulation)]
+    shrink_collinear_vertices(border)
+    return border, holes
+
+
+def _edge_key(edge: QuadEdge) -> Sortable:
+    return _to_squared_edge_length(edge), edge.start, edge.end
+
+
+def _to_squared_edge_length(edge: QuadEdge) -> Coordinate:
+    (start_x, start_y), (end_x, end_y) = edge.start, edge.end
+    delta_x, delta_y = start_x - end_x, start_y - end_y
+    return delta_x * delta_x + delta_y * delta_y
+
+
+def constrict_convex_hull_size(points: List[Point],
+                               *,
+                               max_size: Optional[int]) -> List[Point]:
+    if max_size is None:
+        return points
+    convex_hull = to_strict_convex_hull(points)
+    if len(convex_hull) <= max_size:
+        return points
+    sorted_convex_hull = sorted(
+            convex_hull,
+            key=partial(_to_squared_points_distance, convex_hull[0]))
+    new_border_points = []
+    for index in range(max_size):
+        quotient, remainder = divmod(index, 2)
+        if remainder:
+            new_border_points.append(sorted_convex_hull[-quotient - 1])
+        else:
+            new_border_points.append(sorted_convex_hull[quotient])
+    new_border = to_convex_hull(new_border_points)
+    new_border_extra_segments = tuple(
+            {(new_border[index - 1], new_border[index])
+             for index in range(len(new_border))}
+            - {(convex_hull[index], convex_hull[index - 1])
+               for index in range(len(convex_hull))})
+    return (new_border
+            + [point
+               for point in set(points) - set(convex_hull)
+               if all(to_orientation(end, start, point)
+                      is Orientation.COUNTERCLOCKWISE
+                      for start, end in new_border_extra_segments)])
+
+
+def _to_squared_points_distance(left: Point, right: Point) -> Coordinate:
+    (left_x, left_y), (right_x, right_y) = left, right
+    return (left_x - right_x) ** 2 + (left_y - right_y) ** 2
+
+
 def to_convex_contour(points: List[Point],
                       x_flags: List[bool],
                       y_flags: List[bool],
@@ -121,11 +203,9 @@ def to_convex_contour(points: List[Point],
         min_polygon_x, min_polygon_y = (min(min_polygon_x, point_x),
                                         min(min_polygon_y, point_y))
     shift_x, shift_y = min_x - min_polygon_x, min_y - min_polygon_y
-    contour = [(min(max(point_x + shift_x, min_x), max_x),
-                min(max(point_y + shift_y, min_y), max_y))
-               for point_x, point_y in points]
-    shrink_collinear_vertices(contour)
-    return to_convex_hull(contour)
+    return to_strict_convex_hull([(min(max(point_x + shift_x, min_x), max_x),
+                                   min(max(point_y + shift_y, min_y), max_y))
+                                  for point_x, point_y in points])
 
 
 def shrink_collinear_vertices(contour: Contour) -> None:
@@ -154,6 +234,28 @@ def to_convex_hull(points: Sequence[Point]) -> Contour:
 
 
 def _to_sub_hull(points: Iterable[Point]) -> List[Point]:
+    result = []
+    for point in points:
+        while len(result) >= 2:
+            if (to_orientation(result[-1], result[-2], point)
+                    is Orientation.CLOCKWISE):
+                del result[-1]
+            else:
+                break
+        result.append(point)
+    return result
+
+
+def to_strict_convex_hull(points: Sequence[Point]) -> Contour:
+    points = sorted(points)
+    lower = _to_strict_sub_hull(points)
+    upper = _to_strict_sub_hull(reversed(points))
+    result = lower[:-1] + upper[:-1]
+    shrink_collinear_vertices(result)
+    return result
+
+
+def _to_strict_sub_hull(points: Iterable[Point]) -> List[Point]:
     result = []
     for point in points:
         while len(result) >= 2:
